(* Opgave 8.6 - calculator *)
load "Math";

type Stack = real list;
datatype Instruction = ++ | -- | ** | // | SIN | COS | LN | EXP | PUSH of real;
exception InvalidInst of string;
(* calc: Instruction * Stack -> Stack *)
local
  fun binary f ([]) = raise InvalidInst "Must supply two operands to binary operator"
   |  binary f (x::[]) = raise InvalidInst "Must supply two operands to binary operator"
   |  binary f (x::y::l) = f(x,y)::l;
  fun unary f ([]) = raise InvalidInst "Must supply at least one operand to unary operator"
   |  unary f (x::xl) = f(x)::xl;
in
fun calc( i: Instruction, s ) = 
  case i of
    ++   => binary op+ (s)
  | --   => binary op- (s)
  | **   => binary op* (s)
  | //   => binary op/ (s)
  | SIN  => unary Math.sin (s)
  | COS  => unary Math.cos (s)
  | LN   => unary Math.ln (s)
  | EXP  => unary Math.exp (s)
  | PUSH a => a::s
  | _ => raise InvalidInst "Invalid operation requested"
 end


(* calcProg: Instruction list -> real *)
fun calcProg( [] ) = NONE
 |  calcProg( xl ) = SOME(  hd(foldl (fn (a,s)=>calc(a,s)) [] (xl)) );

val s = calc(PUSH 10.0, []);
val s = calc(EXP, s);
val p = calcProg( [PUSH 10.0, EXP] );


(* fexpr *)
infix 6 +++ ---;
infix 7 *** ///;
datatype fexpr =
    Const of real
  | X
  | +++ of fexpr * fexpr | --- of fexpr * fexpr
  | *** of fexpr * fexpr | /// of fexpr * fexpr
  | Sin of fexpr | Cos of fexpr
  | Ln of fexpr | Exp of fexpr;


(* toInst: fexpr * real -> Instruction list *)
fun toInst( Const a, _ ) = [PUSH a]
 |  toInst( X, x ) = [PUSH x]
 |  toInst( a +++ b, x) = toInst(a,x) @ toInst(b,x) @ [++]
 |  toInst( a --- b, x) = toInst(a,x) @ toInst(b,x) @ [--]
 |  toInst( a *** b, x) = toInst(a,x) @ toInst(b,x) @ [**]
 |  toInst( a /// b, x) = toInst(a,x) @ toInst(b,x) @ [//]
 |  toInst( Ln(a), x ) = toInst(a,x) @ [LN]
 |  toInst( Exp(a), x ) = toInst(a,x) @ [EXP]
 |  toInst( Sin(a), x ) = toInst(a,x) @ [SIN]
 |  toInst( Cos(a), x ) = toInst(a,x) @ [COS];


(* calcExpr: fexpr * real -> real *)
fun calcExpr( expr, x ) = calcProg( toInst( expr, x ) );

(* Scan: string -> token list *)
fun Scan( s ) = String.tokens Char.isSpace (s);

(* Parse: token list -> fexpr *)
(* Search for (ignore all in paranthesis) plus in expression, if found - part string at "+" and parse each substring.
 * if no plus is found, search for "-". Divide at minus and parse
 *)












