(* Datatypes: Chapt. 7*)
load "Math";
load "Real";
load "Int";

(* Opgave 7.1 *)
datatype Roots = COMPLEX | ONE of real | TWO of (real*real);
datatype Polynome = Pol of (real*real*real) | IntPol of (int*int*int);

fun Solve( Pol(0.0,b,c) ) = ONE(~c/b)
 |  Solve( Pol(a,b,c) ) = 
    let 
	val d = b*b-4.0*a*c;
    in
	 if (d<0.0) then COMPLEX
	 else 
	      if (d=0.0) then ONE(~b/(2.0*a)) else
	      TWO((~b-Math.sqrt d)/(2.0*a),(~b+Math.sqrt d)/(2.0*a))
    end
 | Solve( IntPol(a,b,c) ) = Solve ( Pol(Real.fromInt a, Real.fromInt b, Real.fromInt c) );


(* Opgave 7.2 *)
(* smallest: int list -> int option *)
fun Smallest( [] ) = NONE
 |  Smallest( x::xs ) = 
    case Smallest(xs) of 
      SOME y => if y<x then SOME y else SOME x
    | NONE   => SOME x;
    


(* Record implementation of time *)
type rTime = { Hour:int, Min:int, Sec: int, ampm: string };

local
    fun toSec( {Hour=hour, Min=min, Sec=sec, ampm=dp } ) = hour*3600+min*60+sec + (fn p=>if (p = "pm") then 12*3600 else 0) (dp);
in
    fun compareTime( t1, t2 ) = toSec(t1) > toSec(t2);
end;
	   

(* Datatype implementation of time *)
datatype Time = TOD of int*int*int*string | TOD24 of int*int*int | TODhm of int*int*string | TODhm24 of int*int;
 
local
    fun toSec( TOD24(hour,min,sec) ) = hour*3600+min*60+sec
     |  toSec( TOD(hour,min,sec,ampm) ) = if (ampm = "pm") then toSec( TOD24(hour+12,min,sec) ) else toSec( TOD24(hour,min,sec) )
     |  toSec( TODhm(hour,min,ampm) ) = toSec( TOD(hour,min,0,ampm) )
     |  toSec( TODhm24(hour,min) ) = toSec( TOD24(hour,min,0) );
in
    fun compare( t1, t2 ) = Int.compare(toSec(t1),toSec(t2));
end;




(* Abstype implementation of Grades *)

abstype Marks = Passed of bool | Grade of int 
with

exception EInvalidGrade;
exception EMissingGrades;

fun mkPassed t = Passed t;
fun mkGrade g = if (g=1) orelse (g=2) orelse (g=4) orelse (g=12) orelse (g>13) orelse (g<0) then raise EInvalidGrade else Grade g;

fun didpass( Passed(x) ) = x
 |  didpass( Grade(x) ) = x>5;

local 
    fun calcAverage( [], 0, _) = raise EMissingGrades
     |  calcAverage( [], n, s ) = Real.fromInt(s) / Real.fromInt(n)
     |  calcAverage( Passed(_)::xs, n, s ) = calcAverage(xs,n,s)
     |  calcAverage( Grade(x)::xs, n, s ) = calcAverage(xs,n+1,s+x);
in
    fun average( xl ) = calcAverage(xl,0,0);
end

fun Average( xl ) = Real.fromInt(foldl (fn (x,p) => case x of Grade(g) => g+p | _ => p) 0 ( xl )) / Real.fromInt( foldl (fn (x,n) => case x of Grade(_) => n+1 | _ => n) 0 (xl));

local 
    fun countp( [], n ) = n
     |  countp( x::xl, n ) = if didpass(x) then countp(xl,n+1) else countp(xl,n)
in
    fun countPassed( xl ) = countp( xl,0 );
end

fun countp( [] ) = 0
 |  countp( x::xl ) = if didpass(x) then 1+countp(xl) else countp(xl);

fun countPassedHO( xl ) = foldl (fn (x,n) => if didpass(x) then n+1 else n) 0 (xl);


end


val marks = [ mkPassed true, mkGrade 10, mkGrade 13, mkGrade 0, mkPassed false, mkGrade 3];








