(* Treeeeeeeeeeeeess!!! *)
load "Real";

(* fexpr *)
infix 6 +++ ---;
infix 7 *** ///;
datatype fexpr =
    Const of real
  | X
  | +++ of fexpr * fexpr | --- of fexpr * fexpr
  | *** of fexpr * fexpr | /// of fexpr * fexpr
  | Sin of fexpr | Cos of fexpr
  | Ln of fexpr | Exp of fexpr;

fun postfix( exp: fexpr ) : string = 
    case exp of
	Const(a) => Real.toString(a)
    |   X => "x"
    |   op+++(exp1,exp2) => "("^postfix(exp1)^","^postfix(exp2)^")+"
    |   op---(exp1,exp2) => "("^postfix(exp1)^","^postfix(exp2)^")-"
    |   op***(exp1,exp2) => "("^postfix(exp1)^","^postfix(exp2)^")*"
    |   op///(exp1,exp2) => "("^postfix(exp1)^","^postfix(exp2)^")/"
    |   Sin(exp1) => "("^postfix(exp1)^")Sin"
    |   Cos(exp1) => "("^postfix(exp1)^")Cos"
    |   Ln (exp1) => "("^postfix(exp1)^")Ln"
    |   Exp(exp1) => "("^postfix(exp1)^")Exp";

(* 8.4 - Proposition logix *)
(* a *)
infix 7 AND;
infix 6 OR;

datatype logic = 
  ATOM of string
| NOT of logic
| AND of logic*logic
| OR of logic*logic;

(* b *)

fun deMorgan ( x:logic ):logic =
    case x of
      NOT( NOT(l1) ) => l1  
    | NOT( l1 AND l2 ) =>  NOT(deMorgan(l1)) OR  NOT(deMorgan(l2))
    | NOT( l1 OR  l2 ) =>  NOT(deMorgan(l1)) AND NOT(deMorgan(l2))
    | l1 AND l2 => deMorgan(l1) AND deMorgan(l2)
    | NOT(l1) => deMorgan(l1)
    | l1 OR l2 => deMorgan(l1) OR deMorgan(l2)
    | _ => x;

(* c *)

fun expand( x:logic ):logic = 
    case x of
      l1 OR (l2 AND l3) => (l1 OR l2) AND (l1 OR l3)
    | (l1 AND l2) OR l3 => (l1 OR l2) AND (l2 OR l3)
    | _ => x;

(* d *)

fun tautology( x: logic ): bool =
    case x of
      ATOM(p) OR NOT(ATOM(q)) => p=q
    | NOT(ATOM(p)) OR ATOM(q) => p=q
    | l1 AND l2 => tautology(l1) andalso tautology(l2)
    | l1 OR l2 => tautology(l1) orelse tautology(l2)
    | NOT(l1) => not (tautology(l1))
    | ATOM(_) => false;

fun toString( ATOM(s) ) = s
 |  toString( NOT(l) ) = "!("^toString(l)^")"
 |  toString( l1 AND l2 ) = "( "^toString(l1)^" and "^toString(l2)^" )"
 |  toString( l1 OR l2 ) = "( "^toString(l1)^" or "^toString(l2)^" )";

val log1 = NOT( ATOM("p") AND ATOM("q")) OR NOT( ATOM("p") OR ATOM("q"));


(*Exercise 8.5*)

(*a+b*)

datatype department =
  dep of string*real*department list;

(*c*)
(* extract: department -> (string*real) list *)

fun extract( dep(name,gross,subdeps) ) = 
    (name,gross)::(foldl (fn (a,b)=>b @ extract(a)) [] (subdeps));     


local 

 fun extsubdep( [] : department list ): (string*real) list = []
  |  extsubdep( s::sl : department list ) = extract2(s) @ extsubdep(sl)
 and extract2( dep(name: string, gross: real, s: department list ) : department ) = (name,gross) :: extsubdep(s)

in
    fun Extract(d)=extract2(d);
end

val cokeland = dep("coke-cola",1000.0, [ dep("brew",10.0,[dep("waterplant",5.0,[])]), dep("drugs",1000.0,[dep("refine",100.0,[]),dep("produce",100.0,[]),dep("hero-in",~2.0,[dep("corruption",~200.0,[]),dep("farmers",~20.0,[]),dep("soldiers",~80.0,[dep("armor",~10.0,[]),dep("foot",~50.0,[])])]) ]) ]);


(*e*)
fun getGross( dep(name,gross, subdeps) ) =
    foldl (fn ((_,gr),tot) => tot+gr) 0.0 ( Extract(dep(name,gross,subdeps)) ) + gross;

local 
    fun extsubdep( [] ) = []
     |  extsubdep( s::sl ) = extractdep(s) @ extsubdep(sl)
    and extractdep( dep(name,gross,s) ) = (name, getGross( dep(name,gross,s) ) ) :: extsubdep(s)
in
    fun getTotals(d) = extractdep(d)
end;

(*f*)
local
    fun extsubdep( [], _ ) = ""
     |  extsubdep( s::sl, indent ) = deptostr(s, indent) ^ extsubdep(sl, indent)
    and deptostr( dep(name,gross,s), indent ) = indent^name^" ("^Real.toString(gross)^")\n"^extsubdep(s, indent^"\t")
in
    fun toString(d) = deptostr(d, "")
end;


(* Higher-order functions *)










